#region Using directives

using System;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.Collections;

#endregion

namespace BitmapHandler
{
    public unsafe class BitmapHandler : IDisposable
    {
        private Bitmap bmp; // the forms bitmap
        Color bcol; //background color
        private BitmapData bmpdata;
        byte* bptr;
        byte* pixel;
        int stride;
        private Random rnd;
        private int btop; // top y of drawn area
        private int bleft; // top x of drawn area
        private int bbottom; // bottom y of drawn area
        private int bright; // bottom x of drawn area

        private int[] x;
        private int[] y;
        private int[] col;
        private double[] spread;
        private double[] gravity;

        //interactions 
        private int[] colors;
        private int numcolors;
        private Hashtable hashcolors;
        private int[] rules;
        private int numrules;
        private double[] ruleprobs;
        private int[] change1;
        private int[] change2;
       

       // constructor
        public BitmapHandler(Bitmap bm, Color backcol, int top, int left, int bottom, int right,
            int[] px, int[] py, int[] pcol, double[] pspread, double[] pgravity)
        {
            bmp = bm;
            rnd = new Random();
            btop = top;
            bleft = left;
            bbottom = bottom;
            bright = right;
            bcol = backcol;
            x = px;
            y = py;
            col = pcol;
            spread = pspread;
            gravity = pgravity;

        }

        // private methods
        private void setcol(int x, int y, int col)
        {
            pixel = bptr + y * stride + x * 3;
            *pixel++ = (byte)col;
            *pixel++ = (byte)(col >> 8);
            *pixel = (byte)(col >> 16);
        }

        private int getcol(int x, int y)
        {
            pixel = bptr + y * stride + x * 3 + 2;
            int col = 0xff;
            col = (col << 8) + *pixel--;
            col = (col << 8) + *pixel--;
            col = (col << 8) + *pixel;
            return col;
        }

        // public properties
        public Bitmap ControlRef
        { get { return bmp; } }

        public double DllVersion
        { get { return 2.21; } }

        public Bitmap Value
        { get { return bmp; } }

        // public methods
        public int AddElement( int n, int pnum, int pen, int px, int py, Color pcol, double pspread, double pgravity)
        {
            for (int i = 0 ; i < pnum ; i++)
            {
                x[n] = Math.Min(Math.Max(px + (int)(rnd.NextDouble() * pen * 2 - pen), bleft), bright);
                y[n] = Math.Min(Math.Max(py + (int)(rnd.NextDouble() * pen * 2 - pen), btop), bbottom);
                col[n] = pcol.ToArgb();
                spread[n] = pspread;
                gravity[n] = pgravity;
                if (bmp.GetPixel(x[n], y[n]) == bcol)
                {
                    n++;
                }
            }
            return n;
        }

        public void InitialiseRules(int nocols, int[] cols, int norules, int[] rles, double[] ruleprbs,
            int[] chnge1, int[]chnge2)
        {
            numcolors = nocols;
            colors = new int[numcolors];
            hashcolors = new Hashtable();
            rules = new int[numcolors * numcolors];
            ruleprobs = new double[numcolors * numcolors];
            change1 = new int[numcolors * numcolors];
            change2 = new int[numcolors * numcolors];
            numrules = norules;

            //get and save the actual colours used on the platform
            Color Saved = bmp.GetPixel(0,0);
            for (int i = 0; i < numcolors; i++)
            {
                bmp.SetPixel(0, 0, Color.FromArgb(cols[i]));
                colors[i]= (bmp.GetPixel(0, 0)).ToArgb();
                hashcolors.Add(colors[i], i);
            }
            bmp.SetPixel(0, 0, Saved);

            // intialise the rules arrays
            for (int i = 0; i < numcolors*numcolors; i++)
            {
                ruleprobs[i] = 0;
                change1[i] = 0;
                change2[i] = 0;
            }

            // add the rules
            for (int i = 0; i < numrules; i++)
            {
                int j = rules[i] % numcolors;
                int k = rules[i] / numcolors;
                // make the rule commutative
                ruleprobs[j * numcolors + k] = ruleprbs[i];
                change1[j * numcolors + k] = chnge1[i];
                change2[j * numcolors + k] = chnge2[i];
                ruleprobs[k * numcolors + j] = ruleprbs[i];
                change1[k * numcolors + j] = chnge1[i];
                change2[k * numcolors + j] = chnge2[i];
            }
        }

        //public int timer(int[] x, int[] y, int[] col, double[] spread, double[] gravity, int n, Color bcol)
        public int timer(int n)
        {
            int i, nx, ny, pcol, hitcolor;
            double rand, grav;
            // unlike the full framework the compact framework will only lock the entire bitmap
            bmpdata = bmp.LockBits(new Rectangle(0, 0, bmp.Width, bmp.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            bptr = (byte*)bmpdata.Scan0;
            stride = bmpdata.Stride;
            for (i = 0; i < n; i++)
            {
                // calculate next position
                nx = Math.Min(Math.Max(x[i] + (int)((2 * rnd.NextDouble() - 1) * spread[i]), bleft), bright);
                ny = y[i];
                rand = rnd.NextDouble();
                grav = gravity[i];
                if ((grav < 0) && (rand < -grav))
                {
                    ny -=  1 ;
                }
                else if ((grav> 0) && (rand < grav))
                {
                    ny += 1;
                }
                ny = Math.Min(Math.Max(ny, btop), bbottom);

                //hitcolor =  (int)hashcols[getcol(nx, ny)];

                // check if we can move downward, stay if not
                if (getcol(x[i], ny) != bcol.ToArgb())
                {
                    ny = y[i];
                }                
                // we may be moving sideways so check that as well
                if (getcol(nx, ny) != bcol.ToArgb())
                {
                    nx = x[i];
                }
                if( (nx != x[i]) || (ny != y[i]) )
                {
                    //undraw old particle position
                    //setcol(x[i], y[i], bcol.ToArgb());
                    pixel = bptr + y[i] * stride + x[i] * 3;
                    pcol = bcol.ToArgb();
                    *pixel++ = (byte)pcol;
                    *pixel++ = (byte)(pcol >> 8);
                    *pixel = (byte)(pcol >> 16);
                    x[i] = nx;
                    y[i] = ny;
                    // draw it or lose it off the bottom
                    if (ny < bbottom && ny > btop)
                    {
                        //setcol(nx,ny,col[i]);
                        pixel = bptr + ny * stride + nx * 3;
                        pcol = col[i];
                        *pixel++ = (byte)pcol;
                        *pixel++ = (byte)(pcol >> 8);
                        *pixel = (byte)(pcol >> 16);

                    }
                    else
                    {
                        x[i] = x[n - 1];
                        y[i] = y[n - 1];
                        col[i] = col[n - 1];
                        spread[i] = spread[n - 1];
                        gravity[i] = gravity[n - 1];
                        n = n - 1;
                    }
                }
            }
            bmp.UnlockBits(bmpdata);
            return n;
        }

        public void Dispose()
        {
            bmp = null;
        }
    }


}
